#!/usr/bin/env python3.2

classifiers = [
"Development Status :: 5 - Production/Stable",
"Environment :: Console",
"Intended Audience :: Developers",
"License :: OSI Approved :: Apache Software License",
"Operating System :: OS Independent",
"Programming Language :: Python :: 2.7",
"Programming Language :: Python :: 3",
"Programming Language :: Python :: 3.2",
"Programming Language :: Python :: 3.3",
"Topic :: Software Development :: Code Generators",
"Topic :: Software Development :: Libraries :: Python Modules",
]

from distutils.core import setup


setup(
   name="3to2",
   packages=["lib3to2","lib3to2.fixes","lib3to2.tests"],
   scripts=["3to2"],
   version="1.1.1",
   url="http://www.startcodon.com/wordpress/?cat=8",
   author="Joe Amenta",
   author_email="airbreather@linux.com",
   classifiers=classifiers,
   description="Refactors valid 3.x syntax into valid 2.x syntax, if a syntactical conversion is possible",
   long_description="",
   license="",
   platforms="",
)



from lib3to2.tests.support import lib3to2FixerTestCase

class Test_bytes(lib3to2FixerTestCase):
    fixer = "bytes"

    def test_bytes_call_1(self):
        b = """bytes(x)"""
        a = """str(x)"""
        self.check(b, a)

    def test_bytes_call_2(self):
        b = """a = bytes(x) + b"florist" """
        a = """a = str(x) + "florist" """
        self.check(b, a)

    def test_bytes_call_noargs(self):
        b = """bytes()"""
        a = """str()"""
        self.check(b, a)

    def test_bytes_call_args_1(self):
        b = """bytes(x, y, z)"""
        a = """str(x).encode(y, z)"""
        self.check(b, a)

    def test_bytes_call_args_2(self):
        b = """bytes(encoding="utf-8", source="dinosaur", errors="dont-care")"""
        a = """str("dinosaur").encode("utf-8", "dont-care")"""
        self.check(b, a)

    def test_bytes_literal_1(self):
        b = '''b"\x41"'''
        a = '''"\x41"'''
        self.check(b, a)

    def test_bytes_literal_2(self):
        b = """b'x'"""
        a = """'x'"""
        self.check(b, a)

    def test_bytes_literal_3(self):
        b = """BR'''\x13'''"""
        a = """R'''\x13'''"""
        self.check(b, a)

    def test_bytes_concatenation(self):
        b = """b'bytes' + b'bytes'"""
        a = """'bytes' + 'bytes'"""
        self.check(b, a)



"""
Fixer for:
map -> itertools.imap
filter -> itertools.ifilter
zip -> itertools.izip
itertools.filterfalse -> itertools.ifilterfalse
"""

from lib2to3 import fixer_base
from lib2to3.pytree import Node
from lib2to3.fixer_util import touch_import, is_probably_builtin

class FixItertools(fixer_base.BaseFix):

    PATTERN = """
              power< names=('map' | 'filter' | 'zip') any*> |
              import_from< 'from' 'itertools' 'import' imports=any > |
              power< 'itertools' trailer< '.' f='filterfalse' any* > > |
              power< f='filterfalse' any* > any*
              """

    def transform(self, node, results):
        syms = self.syms
        imports = results.get("imports")
        f = results.get("f")
        names = results.get("names")
        if imports:
            if imports.type == syms.import_as_name or not imports.children:
                children = [imports]
            else:
                children = imports.children
            for child in children[::2]:
                if isinstance(child, Node):
                    for kid in child.children:
                        if kid.value == "filterfalse":
                            kid.changed()
                            kid.value = "ifilterfalse"
                            break
                elif child.value == "filterfalse":
                    child.changed()
                    child.value = "ifilterfalse"
                    break
        elif names:
            for name in names:
                if is_probably_builtin(name):
                    name.value = "i" + name.value
                    touch_import("itertools", name.value, node)
        elif f:
            f.changed()
            f.value = "ifilterfalse"



"""
Fixer for:
0b111101101 -> __builtins__.long("111101101", 2)
0o755 -> 0755
"""

import re

from lib2to3.pgen2 import token
from lib2to3 import fixer_base
from lib2to3.pygram import python_symbols as syms
from lib2to3.pytree import Node
from lib2to3.fixer_util import Number, Call, Attr, String, Name, ArgList, Comma

baseMAPPING = {'b':2, 'o':8, 'x':16}

def base(literal):
    """Returns the base of a valid py3k literal."""
    literal = literal.strip()
    # All literals that do not start with 0, or are 1 or more zeros.
    if not literal.startswith("0") or re.match(r"0+$",literal):
        return 10
    elif literal[1] not in "box":
        return 0
    return baseMAPPING[literal[1]]


class FixNumliterals(fixer_base.BaseFix):

    # We need to modify all numeric literals except floats, complex.
    def unmatch(self, node):
        """Don't match complex numbers, floats, or base-10 ints"""
        val = node.value
        for bad in "jJ+-.":
            if bad in val: return bad
        base_ = base(val)
        return base_ == 10 or base_ == 16

    def match(self, node):
        """Match number literals that are not excluded by self.unmatch"""
        return (node.type == token.NUMBER) and not self.unmatch(node)

    def transform(self, node, results):
        """
        Call __builtins__.long() with the value and the base of the value.
        This works because 0b10 is int("10", 2), 0o10 is int("10", 8), etc.
        """
        val = node.value
        base_ = base(val)
        if base_ == 8:
            assert val.strip().startswith("0o") or \
            val.strip().startswith("0O"), "Invalid format for octal literal"
            node.changed()
            node.value = "".join(("0",val[2:]))
        elif base_ == 2:
            assert val.startswith("0") and val[1] in "bB", \
                                           "Invalid format for binary literal"
            # __builtins__.long
            func_name = Node(syms.power, Attr(Name("__builtins__"), \
                             Name("long")))
            # ("...", 2)
            func_args = [String("".join(("\"", val.strip()[2:], "\""))), \
                         Comma(), Number(2, prefix=" ")]
            new_node = Call(func_name, func_args, node.prefix)
            return new_node



from lib3to2.tests.support import lib3to2FixerTestCase

class Test_dctsetcomp(lib3to2FixerTestCase):
    fixer = "dctsetcomp"

    def test_dictcomp_straightforward(self):
        b = "{key:val for (key, val) in tuple_of_stuff}"
        a = "dict((key, val) for (key, val) in tuple_of_stuff)"
        self.check(b, a)

    def test_dictcomp_nestedstuff_noif(self):
        b = "{hashlify(spam):valuate(ham).whatsthis(eggs) for \
             (spam, ham, eggs) in spam_iterator}"
        a = "dict((hashlify(spam), valuate(ham).whatsthis(eggs)) for \
             (spam, ham, eggs) in spam_iterator)"
        self.check(b, a)

    def test_dictcomp_nestedstuff_withif(self):
        b = "{moo:(lambda new: None)(cow) for (moo, cow) in \
            farm_animal['cow'] if has_milk()}"
        a = "dict((moo, (lambda new: None)(cow)) for (moo, cow) in \
            farm_animal['cow'] if has_milk())"
        self.check(b, a)

    def test_setcomps(self):
        """
        setcomp fixer should keep everything inside the same
        and only replace the {} with a set() call on a gencomp
        """
        tests = []
        tests.append("milk.price for milk in find_milk(store)")
        tests.append("compute_nth_prime(generate_complicated_thing(\
            n.value(hashlifier))) for n in my_range_func(1, (how_far+offset))")
        tests.append("compute_nth_prime(generate_complicated_thing(\
            n.value(hashlifier))) for n in my_range_func(1, (how_far+offset))\
            if a==b.spam()")
        for comp in tests:
            b = "{%s}" % comp
            a = "set(%s)" % comp
        self.check(b, a)

    def test_prefixes(self):
        b = "spam = {foo for foo in bar}"
        a = "spam = set(foo for foo in bar)"
        self.check(b, a)

        b = "spam = {foo:bar for (foo, bar) in baz}"
        a = "spam = dict((foo, bar) for (foo, bar) in baz)"
        self.check(b, a)



"""
Fixer for:

def something(self):
    super()

->

def something(self):
    super(self.__class__, self)
"""

from lib2to3 import fixer_base
from ..fixer_util import Node, Leaf, token, syms, Name, Comma, Dot

dot_class = Node(syms.trailer, [Dot(), Name("__class__")])

def get_firstparam(super_node):
    parent = super_node.parent
    while parent.type != syms.funcdef and parent.parent:
        parent = parent.parent

    if parent.type != syms.funcdef:
        # super() called without arguments outside of a funcdef
        return None

    children = parent.children
    assert len(children) > 2
    params = children[2]
    assert params.type == syms.parameters
    if len(params.children) < 3:
        # Function has no parameters, therefore super() makes no sense here...
        return None
    args = params.children[1]
    if args.type == token.NAME:
        return args.value
    elif args.type == syms.typedargslist:
        assert len(args.children) > 0
        if args.children[0].type == token.NAME:
            return args.children[0].value
        else:
            # Probably a '*'
            return None


def get_class_name(super_node):
    parent = super_node.parent
    while parent.type != syms.classdef and parent.parent:
        parent = parent.parent

    if parent.type != syms.classdef:
        # super() called without arguments outside of a classdef
        return None

    children = parent.children
    assert len(children) > 2
    class_name = children[1]
    assert class_name.type == token.NAME
    return class_name.value


def insert_args(name, class_name, rparen):
    parent = rparen.parent
    
    if class_name:
        class_node = Node(syms.power, [Name(class_name)])
    else:
        class_node = Node(syms.power, [Name(name), dot_class.clone()])

    idx = parent.children.index(rparen)
    parent.insert_child(idx, Name(name, prefix=" "))
    parent.insert_child(idx, Comma())
    parent.insert_child(idx, class_node)


class FixSuper(fixer_base.BaseFix):

    PATTERN = "power< 'super' trailer< '(' rparen=')' > any* >"

    def transform(self, node, results):
        param = get_firstparam(node)
        if param is None:
            self.cannot_convert(node, "super() with no arguments must be called inside a function that has at least one parameter")
            return
        class_name = get_class_name(node)
        rparen = results["rparen"]
        insert_args(param, class_name, rparen)



"""
Fixer for Python 3 function parameter syntax
This fixer is rather sensitive to incorrect py3k syntax.
"""

# Note: "relevant" parameters are parameters following the first STAR in the list.

from lib2to3 import fixer_base
from ..fixer_util import token, indentation, suitify, String, Newline, Comma, DoubleStar, Name

_assign_template = "%(name)s = %(kwargs)s['%(name)s']; del %(kwargs)s['%(name)s']"
_if_template = "if '%(name)s' in %(kwargs)s: %(assign)s"
_else_template = "else: %(name)s = %(default)s"
_kwargs_default_name = "_3to2kwargs"

def gen_params(raw_params):
    """
    Generator that yields tuples of (name, default_value) for each parameter in the list
    If no default is given, then it is default_value is None (not Leaf(token.NAME, 'None'))
    """
    assert raw_params[0].type == token.STAR and len(raw_params) > 2
    curr_idx = 2 # the first place a keyword-only parameter name can be is index 2
    max_idx = len(raw_params)
    while curr_idx < max_idx:
        curr_item = raw_params[curr_idx]
        prev_item = curr_item.prev_sibling
        if curr_item.type != token.NAME:
            curr_idx += 1
            continue
        if prev_item is not None and prev_item.type == token.DOUBLESTAR:
            break
        name = curr_item.value
        nxt = curr_item.next_sibling
        if nxt is not None and nxt.type == token.EQUAL:
            default_value = nxt.next_sibling
            curr_idx += 2
        else:
            default_value = None
        yield (name, default_value)
        curr_idx += 1

def remove_params(raw_params, kwargs_default=_kwargs_default_name):
    """
    Removes all keyword-only args from the params list and a bare star, if any.
    Does not add the kwargs dict if needed.
    Returns True if more action is needed, False if not
    (more action is needed if no kwargs dict exists)
    """
    assert raw_params[0].type == token.STAR
    if raw_params[1].type == token.COMMA:
        raw_params[0].remove()
        raw_params[1].remove()
        kw_params = raw_params[2:]
    else:
        kw_params = raw_params[3:]
    for param in kw_params:
        if param.type != token.DOUBLESTAR:
            param.remove()
        else:
            return False
    else:
        return True
    
def needs_fixing(raw_params, kwargs_default=_kwargs_default_name):
    """
    Returns string with the name of the kwargs dict if the params after the first star need fixing
    Otherwise returns empty string
    """
    found_kwargs = False
    needs_fix = False

    for t in raw_params[2:]:
        if t.type == token.COMMA:
            # Commas are irrelevant at this stage.
            continue
        elif t.type == token.NAME and not found_kwargs:
            # Keyword-only argument: definitely need to fix.
            needs_fix = True
        elif t.type == token.NAME and found_kwargs:
            # Return 'foobar' of **foobar, if needed.
            return t.value if needs_fix else ''
        elif t.type == token.DOUBLESTAR:
            # Found either '*' from **foobar.
            found_kwargs = True
    else:
        # Never found **foobar.  Return a synthetic name, if needed.
        return kwargs_default if needs_fix else ''

class FixKwargs(fixer_base.BaseFix):

    run_order = 7 # Run after function annotations are removed

    PATTERN = "funcdef< 'def' NAME parameters< '(' arglist=typedargslist< params=any* > ')' > ':' suite=any >"

    def transform(self, node, results):
        params_rawlist = results["params"]
        for i, item in enumerate(params_rawlist):
            if item.type == token.STAR:
                params_rawlist = params_rawlist[i:]
                break
        else:
            return
        # params is guaranteed to be a list starting with *.
        # if fixing is needed, there will be at least 3 items in this list:
        # [STAR, COMMA, NAME] is the minimum that we need to worry about.
        new_kwargs = needs_fixing(params_rawlist)
        # new_kwargs is the name of the kwargs dictionary.
        if not new_kwargs:
            return
        suitify(node)

        # At this point, params_rawlist is guaranteed to be a list
        # beginning with a star that includes at least one keyword-only param
        # e.g., [STAR, NAME, COMMA, NAME, COMMA, DOUBLESTAR, NAME] or
        # [STAR, COMMA, NAME], or [STAR, COMMA, NAME, COMMA, DOUBLESTAR, NAME]

        # Anatomy of a funcdef: ['def', 'name', parameters, ':', suite]
        # Anatomy of that suite: [NEWLINE, INDENT, first_stmt, all_other_stmts]
        # We need to insert our new stuff before the first_stmt and change the
        # first_stmt's prefix.

        suite = node.children[4]
        first_stmt = suite.children[2]
        ident = indentation(first_stmt)

        for name, default_value in gen_params(params_rawlist):
            if default_value is None:
                suite.insert_child(2, Newline())
                suite.insert_child(2, String(_assign_template %{'name':name, 'kwargs':new_kwargs}, prefix=ident))
            else:
                suite.insert_child(2, Newline())
                suite.insert_child(2, String(_else_template %{'name':name, 'default':default_value}, prefix=ident))
                suite.insert_child(2, Newline())
                suite.insert_child(2, String(_if_template %{'assign':_assign_template %{'name':name, 'kwargs':new_kwargs}, 'name':name, 'kwargs':new_kwargs}, prefix=ident))
        first_stmt.prefix = ident
        suite.children[2].prefix = ""

        # Now, we need to fix up the list of params.

        must_add_kwargs = remove_params(params_rawlist)
        if must_add_kwargs:
            arglist = results['arglist']
            if len(arglist.children) > 0 and arglist.children[-1].type != token.COMMA:
                arglist.append_child(Comma())
            arglist.append_child(DoubleStar(prefix=" "))
            arglist.append_child(Name(new_kwargs))
            



"""
Fixer for dictcomp and setcomp:
{foo comp_for} -> set((foo comp_for))
{foo:bar comp_for} -> dict(((foo, bar) comp_for))"""

from lib2to3 import fixer_base
from lib2to3.pytree import Node, Leaf
from lib2to3.pygram import python_symbols as syms
from lib2to3.pgen2 import token
from lib2to3.fixer_util import parenthesize, Name, Call, LParen, RParen

from ..fixer_util import commatize

def tup(args):
    return parenthesize(Node(syms.testlist_gexp, commatize(args)))

class FixDctsetcomp(fixer_base.BaseFix):

    PATTERN = """atom< '{' dictsetmaker< 
                  n1=any [col=':' n2=any]
                    comp_for=comp_for< 'for' any 'in' any [comp_if<'if' any>] >
                  > '}' >"""

    def transform(self, node, results):
        comp_for = results.get("comp_for").clone()
        is_dict = bool(results.get("col")) # is it a dict?
        n1 = results.get("n1").clone()
        if is_dict:
            n2 = results.get("n2").clone()
            n2.prefix = " "
            impl_assign = tup((n1, n2))
        else:
            impl_assign = n1
        our_gencomp = Node(syms.listmaker, [(impl_assign),(comp_for)])
        if is_dict:
            new_node = Node(syms.power, [Name("dict"),
                       parenthesize(Node(syms.atom, [our_gencomp]))])
        else:
            new_node = Node(syms.power, [Name("set"),
                       parenthesize(Node(syms.atom, [our_gencomp]))])
        new_node.prefix = node.prefix
        return new_node
